import saber.assign
import saber.calibrate
import saber.cluster
import saber.gis
import saber.io
import saber.validate

__all__ = [
    'io', 'cluster', 'assign', 'gis', 'calibrate', 'validate',
]

__author__ = 'Riley C. Hales'
__version__ = '0.7.0'
__license__ = 'BSD 3 Clause Clear'



import logging
from multiprocessing import Pool

import numpy as np
import pandas as pd

from .io import asgn_gid_col
from .io import asgn_mid_col
from .io import down_mid_col
from .io import gid_col
from .io import mid_col
from .io import order_col
from .io import read_table
from .io import reason_col
from .io import write_table
from .io import clbl_col
from .io import x_col
from .io import y_col

__all__ = ['mp_assign_all', 'generate', 'assign_gauged', 'map_propagate', 'map_resolve_props', 'map_assign_ungauged', ]

logger = logging.getLogger(__name__)


def mp_assign_all(workdir: str, assign_df: pd.DataFrame, n_processes: int or None = None) -> pd.DataFrame:
    """

    Args:
        workdir: path to the working directory
        assign_df: the assignments table dataframe with the clustering labels already applied
        n_processes: number of processes to use for multiprocessing, passed to Pool

    Returns:
        pd.DataFrame
    """
    gauged_mids = assign_df[assign_df[gid_col].notna()][mid_col].values

    # assign gauged basins
    logger.info('Assigning Gauged Basins')
    assign_df = assign_gauged(assign_df)

    logger.info('Assigning by Hydraulic Connectivity')
    with Pool(n_processes) as p:
        logger.info('Finding Downstream Assignments')
        df_prop_down = pd.concat(p.starmap(map_propagate, [(assign_df, x, 'down') for x in gauged_mids]))
        logger.info('Caching Downstream Assignments')
        write_table(df_prop_down, workdir, 'prop_downstream')

        logger.info('Finding Upstream Assignments')
        df_prop_up = pd.concat(p.starmap(map_propagate, [(assign_df, x, 'up') for x in gauged_mids]))
        logger.info('Caching Upstream Assignments')
        write_table(df_prop_up, workdir, 'prop_upstream')

        logger.info('Resolving Propagation Assignments')
        df_prop = pd.concat([df_prop_down, df_prop_up]).reset_index(drop=True)
        df_prop = pd.concat(p.starmap(map_resolve_props, [(df_prop, x) for x in df_prop[mid_col].unique()]))
        logger.info('Caching Propagation Assignments')
        write_table(df_prop, workdir, 'prop_resolved')

        logger.info('Assign Remaining Basins by Cluster, Spatial, and Physical Decisions')
        for cluster_number in range(assign_df[clbl_col].max() + 1):
            logger.info(f'Assigning basins in cluster {cluster_number}')
            # limit by cluster number
            c_df = assign_df[assign_df[clbl_col] == cluster_number]
            # keep a list of the unassigned basins in the cluster
            mids = c_df[c_df[reason_col] == 'unassigned'][mid_col].values
            # filter cluster dataframe to find only gauged basins
            c_df = c_df[c_df[gid_col].notna()]
            assign_df = pd.concat([
                pd.concat(p.starmap(map_assign_ungauged, [(assign_df, c_df, x) for x in mids])),
                assign_df[~assign_df[mid_col].isin(mids)]
            ]).reset_index(drop=True)

    logger.info('SABER Assignment Analysis Completed')
    return assign_df


def generate(workdir: str, labels_df: pd.DataFrame = None, drain_table: pd.DataFrame = None,
             gauge_table: pd.DataFrame = None, cache: bool = True) -> pd.DataFrame:
    """
    Joins the drain_table.csv and gauge_table.csv to create the assign_table.csv

    Args:
        workdir: path to the working directory
        cache: whether to cache the assign table immediately
        labels_df: a dataframe with a column for the assigned cluster label and a column for the model_id
        drain_table: the drain table dataframe
        gauge_table: the gauge table dataframe

    Returns:
        pd.DataFrame
    """
    # read the tables if they are not provided
    if labels_df is None:
        labels_df = read_table(workdir, 'cluster_labels')
    if drain_table is None:
        drain_table = read_table(workdir, 'drain_table')
    if gauge_table is None:
        gauge_table = read_table(workdir, 'gauge_table')

    # enforce correct column data types
    labels_df[mid_col] = labels_df[mid_col].astype(str)
    drain_table[mid_col] = drain_table[mid_col].astype(str)
    drain_table[down_mid_col] = drain_table[down_mid_col].astype(str)
    gauge_table[mid_col] = gauge_table[mid_col].astype(str)

    # join the drain_table and gauge_table then join the labels_df
    assign_df = pd.merge(
        drain_table,
        gauge_table,
        on=mid_col,
        how='outer'
    ).merge(labels_df, on=mid_col, how='outer')
    assign_df = assign_df.sort_values(by=mid_col).reset_index(drop=True)

    # create new columns asgn_mid_col, asgn_gid_col, reason_col
    assign_df[[asgn_mid_col, asgn_gid_col, reason_col]] = ['unassigned', 'unassigned', 'unassigned']

    if cache:
        write_table(assign_df, workdir, 'assign_table')
        write_table(assign_df[[mid_col, ]], workdir, 'mid_list')
        write_table(assign_df[[gid_col, ]], workdir, 'gid_list')
        write_table(assign_df[[mid_col, gid_col]], workdir, 'mid_gid_map')

    return assign_df


def assign_gauged(df: pd.DataFrame) -> pd.DataFrame:
    """
    Assigns basins a gauge for correction which contain a gauge

    Args:
        df: the assignments table dataframe

    Returns:
        Copy of df1 with assignments made
    """
    selector = df[gid_col].notna()
    df.loc[selector, asgn_mid_col] = df[mid_col]
    df.loc[selector, asgn_gid_col] = df[gid_col]
    df.loc[selector, reason_col] = 'gauged'
    return df


def map_propagate(df: pd.DataFrame, start_mid: int, direction: str) -> pd.DataFrame or None:
    """
    Meant to be mapped over a dataframe to propagate assignments downstream or upstream

    Args:
        df: the assignments table dataframe
        start_mid: the model_id to start the propagation from
        direction: either 'down' or 'up' to indicate the direction of propagation

    Returns:
        pd.DataFrame
    """
    assigned_rows = []
    start_order = df[df[mid_col] == start_mid][order_col].values[0]

    # select the starting row
    stream_row = df[df[mid_col] == start_mid]
    start_gid = stream_row[asgn_gid_col].values[0]
    select_same_order_streams = df[order_col] == start_order

    n_steps = 1

    # repeat as long as the current row is not empty
    try:
        while True:
            # select the next up or downstream
            if direction == 'down':
                id_selector = df[mid_col] == stream_row[down_mid_col].values[0]
            else:  # direction == 'up':
                id_selector = df[down_mid_col] == stream_row[mid_col].values[0]

            # select the next row using the ID and Order selectors
            stream_row = df[np.logical_and(id_selector, select_same_order_streams)]

            # Break the loop if
            # 1. next row is empty -> no upstream/downstream row -> empty stream_row
            # 2. next row stream order not a match -> not picked by select_same_order_streams -> empty stream_row
            # 3. next row already has an assignment made -> reason column is not empty string
            if stream_row.empty or stream_row[reason_col].values[0] != 'unassigned':
                break

            # copy the row, modify the assignment columns, and append to the list
            new_row = stream_row.copy()
            new_row[[asgn_mid_col, asgn_gid_col, reason_col]] = [start_mid, start_gid, f'prop-{direction}-{n_steps}']
            assigned_rows.append(new_row)

            # increment the steps counter
            n_steps = n_steps + 1

            # Break the loop if
            # 1. The next row is an outlet -> no downstream row -> cause error when selecting next row
            # 2. we have reach the max number of steps (n_steps -1)
            if stream_row[down_mid_col].values[0] != -1 or n_steps >= 8:
                break
    except Exception as e:
        logger.error(f'Error in map_propagate: {e}')

    if len(assigned_rows):
        return pd.concat(assigned_rows)
    return pd.DataFrame(columns=df.columns)


def map_resolve_props(df_props: pd.DataFrame, mid: str) -> pd.DataFrame:
    """
    Resolves the propagation assignments by choosing the assignment with the fewest steps

    Args:
        df_props: the combined upstream and downstream propagation assignments dataframe
        mid: the model_id to resolve the propagation assignments for

    Returns:
        pd.DataFrame
    """
    df_mid = df_props[df_props[mid_col] == mid].copy()
    # parse the reason statement into number of steps and prop up or downstream
    df_mid[['direction', 'n_steps']] = pd.DataFrame(
        df_props[reason_col].apply(lambda x: x.split('-')[1:]).to_list())
    df_mid['n_steps'] = df_mid['n_steps'].astype(float)
    # sort by n_steps then by reason
    df_mid = df_mid.sort_values(['n_steps', 'direction'], ascending=[True, True])
    # return the first row which is the fewest steps and preferring downstream to upstream
    return df_mid.head(1).drop(columns=['direction', 'n_steps'])


def assign_propagation(df: pd.DataFrame, df_props: pd.DataFrame) -> pd.DataFrame:
    """
     Merge the assignment table and the propagation assignments

    Args:
        df: the assignments table dataframe
        df_props: the combined upstream and downstream propagation assignments dataframe

    Returns:
        pd.DataFrame
    """
    return pd.concat([df[~df[mid_col].isin(df_props[mid_col])], df_props])


def map_assign_ungauged(assign_df: pd.DataFrame, gauges_df: np.array, mid: str) -> pd.DataFrame:
    """
    Assigns all possible ungauged basins a gauge that is
        (1) is closer than any other gauge
        (2) is of same stream order as ungauged basin
        (3) in the same simulated fdc cluster as ungauged basin

    Args:
        assign_df: the assignments table dataframe
        gauges_df: a subset of the assignments dataframe containing the gauges
        mid: the model_id to assign a gauge for

    Returns:
        a new row for the given mid with the assignments made
    """
    try:
        # todo account for physical parameters if they are included
        # find the closest gauge using euclidean distance without accounting for projection/map distortion
        mid_x, mid_y = assign_df.loc[assign_df[mid_col] == mid, [x_col, y_col]].head(1).values.flatten()
        row_idx_to_assign = pd.Series(
            np.sqrt(np.power(gauges_df[x_col] - mid_x, 2) + np.power(gauges_df[y_col] - mid_y, 2))
        ).idxmin()

        asgn_mid, asgn_gid = gauges_df.loc[row_idx_to_assign, [mid_col, gid_col]]
        asgn_reason = f'cluster-{gauges_df[clbl_col].values[0]}'

        new_row = assign_df[assign_df[mid_col] == mid].copy()
        new_row[[asgn_mid_col, asgn_gid_col, reason_col]] = [asgn_mid, asgn_gid, asgn_reason]
    except Exception as e:
        logger.error(f'Error in map_assign_ungauged: {e}')
        new_row = pd.DataFrame(columns=assign_df.columns)

    return new_row



import glob
import logging
import math
import os
from collections.abc import Iterable

import joblib
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from kneed import KneeLocator
from natsort import natsorted
from sklearn.cluster import MiniBatchKMeans
from sklearn.metrics import silhouette_samples

from .io import _find_model_files
from .io import clbl_col
from .io import mid_col
from .io import get_table_path
from .io import read_table
from .io import write_table

__all__ = [
    'cluster',
    'generate',
    'summarize_fit', 'calc_silhouette',
    'plot_fit_metrics', 'plot_centers', 'plot_clusters', 'plot_silhouettes',
    'predict_labels'
]

logger = logging.getLogger(__name__)


def cluster(workdir: str, train_df: str) -> None:
    """
    Train k-means cluster models, calculate fit metrics, and generate plots

    Args:
        workdir: path to the project directory
        train_df: path to the prepared FDC data for training cluster models

    Returns:
        None
    """
    logger.info('Generate Clusters')
    x_fdc_train = pd.read_parquet(train_df, engine='fastparquet').values
    generate(workdir, x=x_fdc_train)
    summarize_fit(workdir)
    # calc_silhouette(workdir, x=x_fdc_train, n_clusters=range(2, 10))

    logger.info('Create Plots')
    plot_clusters(workdir, x=x_fdc_train)
    plot_centers(workdir)
    plot_fit_metrics(workdir)
    # plot_silhouettes(workdir)
    return


def generate(workdir: str, x: np.ndarray = None, max_clusters: int = 13) -> None:
    """
    Trains scikit-learn MiniBatchKMeans models and saves as pickle

    Args:
        workdir: path to the project directory
        x: a numpy array of the prepared FDC data
        max_clusters: maximum number of clusters to train

    Returns:
        None
    """
    if x is None:
        x = read_table(workdir, 'hindcast_fdc_trans').values

    # build the kmeans model for a range of cluster numbers
    for n_clusters in range(2, max_clusters + 1):
        logger.info(f'Clustering n={n_clusters}')
        kmeans = MiniBatchKMeans(n_clusters=n_clusters, init='k-means++', n_init=100)
        kmeans.fit_predict(x)
        joblib.dump(kmeans, os.path.join(workdir, 'clusters', f'kmeans-{n_clusters}.pickle'))
    return


def predict_labels(workdir: str, n_clusters: int, x: pd.DataFrame) -> pd.DataFrame:
    """
    Predict the cluster labels for a set number of FDCs

    Args:
        workdir: path to the project directory
        n_clusters: number of cluster model to use for prediction
        x: A dataframe with 1 row per FDC (stream) and 1 column per FDC value. Index is the stream's ID.

    Returns:
        None
    """
    model = joblib.load(os.path.join(workdir, 'clusters', f'kmeans-{n_clusters}.pickle'))
    labels_df = pd.DataFrame(
        np.transpose([model.predict(x.values), x.index]),
        columns=[clbl_col, mid_col]
    )
    write_table(labels_df, workdir, 'cluster_labels')
    return labels_df


def summarize_fit(workdir: str) -> None:
    """
    Generate a summary of the clustering results save the centers and labels to parquet

    Args:
        workdir: path to the project directory

    Returns:
        None
    """
    summary = {'number': [], 'inertia': [], 'n_iter': []}
    labels = []

    for model_file in _find_model_files(workdir, n_clusters='all'):
        logger.info(f'Post Processing {os.path.basename(model_file)}')
        kmeans = joblib.load(model_file)
        n_clusters = int(kmeans.n_clusters)
        labels.append(kmeans.labels_.flatten())

        # save cluster centroids to table - columns are the cluster number, rows are the centroid FDC values
        write_table(
            pd.DataFrame(np.transpose(kmeans.cluster_centers_), columns=np.array(range(n_clusters)).astype(str)),
            workdir,
            f'cluster_centers_{n_clusters}'
        )

        # save the summary stats from this model
        summary['number'].append(n_clusters)
        summary['inertia'].append(kmeans.inertia_)
        summary['n_iter'].append(kmeans.n_iter_)

    # save the summary results as a csv
    sum_df = pd.DataFrame(summary)
    sum_df['knee'] = KneeLocator(summary['number'], summary['inertia'], curve='convex', direction='decreasing').knee
    write_table(sum_df, workdir, 'cluster_metrics')

    return


def calc_silhouette(workdir: str, x: np.ndarray, n_clusters: int or Iterable = 'all', samples: int = 75_000) -> None:
    """
    Calculate the silhouette score for the given number of clusters

    Args:
        workdir: path to the project directory
        x: a numpy array of the prepared FDC data
        n_clusters: the number of clusters to calculate the silhouette score for
        samples: the number of samples to use for the silhouette score calculation

    Returns:
        None
    """
    if x is None:
        x = read_table(workdir, 'hindcast_fdc_trans').values
    fdc_df = pd.DataFrame(x)

    summary = {'number': [], 'silhouette': []}

    random_shuffler = np.random.default_rng()

    for model_file in _find_model_files(workdir, n_clusters):
        logger.info(f'Calculating Silhouettes for {os.path.basename(model_file)}')
        kmeans = joblib.load(model_file)

        # randomly sample fdcs from each cluster
        fdc_df['label'] = kmeans.labels_
        ss_df = pd.DataFrame(columns=fdc_df.columns.to_list())
        for i in range(int(kmeans.n_clusters)):
            values = fdc_df[fdc_df['label'] == i].drop(columns='label').values
            random_shuffler.shuffle(values)
            values = values[:int(samples)]
            tmp = pd.DataFrame(values)
            tmp['label'] = i
            ss_df = pd.concat([ss_df, tmp])

        # calculate their silhouette scores
        ss_df['silhouette'] = silhouette_samples(ss_df.drop(columns='label').values, ss_df['label'].values, n_jobs=-1)
        ss_df['silhouette'] = ss_df['silhouette'].round(3)
        ss_df.columns = ss_df.columns.astype(str)
        write_table(ss_df, workdir, f'cluster_sscores_{kmeans.n_clusters}')

        # save the summary stats from this model
        summary['number'].append(kmeans.n_clusters)
        summary['silhouette'].append(ss_df['silhouette'].mean())

    # save the summary stats
    write_table(pd.DataFrame(summary), workdir, 'cluster_sscores')
    return


def plot_clusters(workdir: str, x: np.ndarray = None, n_clusters: int or Iterable = 'all',
                  max_cols: int = 3, plt_width: int = 2, plt_height: int = 2, n_lines: int = 2_500) -> None:
    """
    Generate figures of the clustered FDC's

    Args:
        workdir: path to the project directory
        x: a numpy array of the prepared FDC data
        n_clusters: number of clusters to create figures for
        max_cols: maximum number of columns (subplots) in the figure
        plt_width: width of each subplot in inches
        plt_height: height of each subplot in inches
        n_lines: max number of lines to plot in each subplot

    Returns:
        None
    """
    if x is None:
        x = read_table(workdir, 'hindcast_fdc_trans').values

    size = x.shape[1]
    x_values = np.linspace(0, size, 5)
    x_ticks = np.linspace(0, 100, 5).astype(int)

    random_shuffler = np.random.default_rng()

    for model_file in _find_model_files(workdir, n_clusters):
        logger.info(f'Plotting Clusters {os.path.basename(model_file)}')

        # load the model and calculate
        kmeans = joblib.load(model_file)
        n_clusters = int(kmeans.n_clusters)
        n_cols = min(n_clusters, max_cols)
        n_rows = math.ceil(n_clusters / n_cols)

        # initialize the figure and labels
        fig, axs = plt.subplots(
            n_rows,
            n_cols,
            figsize=(plt_width * n_cols + 1, plt_height * n_rows + 1),
            dpi=750,
            squeeze=False,
            tight_layout=True,
            sharey='row'
        )
        fig.suptitle("KMeans FDC Clustering")
        fig.supxlabel('Exceedance Probability (%)')
        fig.supylabel('Discharge Z-Score')

        for i, ax in enumerate(fig.axes[:n_clusters]):
            ax.set_title(f'Cluster {i + 1} (n = {np.sum(kmeans.labels_ == i)})')
            ax.set_xlim(0, size)
            ax.set_xticks(x_values, x_ticks)
            ax.set_ylim(-2, 4)
            fdc_sample = x[kmeans.labels_ == i]
            random_shuffler.shuffle(fdc_sample)
            fdc_sample = fdc_sample[:n_lines]
            for j in fdc_sample:
                ax.plot(j.ravel(), "k-")
            ax.plot(kmeans.cluster_centers_[i].flatten(), "r-")
        # turn off plotting axes which are blank (when ax number > n_clusters)
        for ax in fig.axes[n_clusters:]:
            ax.axis('off')

        fig.savefig(os.path.join(workdir, 'clusters', f'figure-clusters-{n_clusters}.png'))
        plt.close(fig)
    return


def plot_silhouettes(workdir: str, plt_width: int = 3, plt_height: int = 3) -> None:
    """
    Plot the silhouette scores for each cluster.
    Based on https://scikit-learn.org/stable/auto_examples/cluster/plot_kmeans_silhouette_analysis.html

    Args:
        workdir: path to the project directory
        plt_width: width of each subplot in inches
        plt_height: height of each subplot in inches

    Returns:
        None
    """
    logger.info('Generating Silhouette Diagrams')

    clusters_dir = os.path.join(workdir, 'clusters')

    for sscore_table in natsorted(glob.glob(os.path.join(clusters_dir, 'cluster_sscores_*.parquet'))):
        logger.info(f'Generating Silhouette Diagram: {os.path.basename(sscore_table)}')
        n_clusters = int(sscore_table.split('_')[-1].split('.')[0])
        sscore_df = pd.read_parquet(sscore_table, engine='fastparquet')
        centers_df = read_table(workdir, f'cluster_centers_{n_clusters}')
        mean_ss = sscore_df['silhouette'].mean()

        # initialize the figure
        fig, (ax1, ax2) = plt.subplots(
            nrows=1,
            ncols=2,
            figsize=(plt_width * 2 + 1, plt_height + 1),
            dpi=600,
            tight_layout=True,
        )

        # Plot 1 titles and labels
        ax1.set_title(f"Silhouette Plot (mean={mean_ss:.3f})")
        ax1.set_xlabel("Silhouette Score")
        ax1.set_ylabel("Cluster Label")
        ax1.set_yticks([])  # Clear the yaxis labels / ticks
        ax1.set_xticks([-0.2, 0, 0.2, 0.4, 0.6, 0.8, 1])

        # Plot 2 titles and labels
        ax2.set_title("Cluster Centers")
        ax2.set_xlabel("Exceedance Probability (%)")
        ax2.set_ylabel("Discharge Z-Score")

        # The vertical line for average silhouette score of all the values
        ax1.axvline(x=mean_ss, color="red", linestyle="--")

        y_lower = 10
        for sub_cluster in range(int(n_clusters)):
            # select the rows applicable to the current sub cluster
            cluster_sscores = sscore_df[sscore_df['label'] == sub_cluster]['silhouette'].values.flatten()
            cluster_sscores.sort()

            n = cluster_sscores.shape[0]
            y_upper = y_lower + n

            color = cm.nipy_spectral(sub_cluster / int(n_clusters))
            ax1.fill_betweenx(
                np.arange(y_lower, y_upper),
                0,
                cluster_sscores,
                facecolor=color,
                edgecolor=color,
                alpha=0.7,
            )
            # Label the silhouette plots with their cluster numbers at the middle
            ax1.text(-0.05, y_lower + 0.5 * n, f'{sub_cluster + 1}: n={n:,}')

            # plot the cluster center
            ax2.plot(centers_df[f'{sub_cluster}'].values, alpha=0.7, c=color, label=f'Cluster {sub_cluster + 1}')

            # add some buffer before the next cluster
            y_lower = y_upper + 10

        fig.savefig(os.path.join(clusters_dir, f'figure-silhouette-diagram-{n_clusters}.png'))
        plt.close(fig)
    return


def plot_centers(workdir: str, plt_width: int = 2, plt_height: int = 2, max_cols: int = 3) -> None:
    """
    Plot the cluster centers for each cluster.

    Args:
        workdir: path to the project directory
        plt_width: width of each subplot in inches
        plt_height: height of each subplot in inches
        max_cols: maximum number of columns of subplots in the figure

    Returns:
        None
    """
    logger.info('Plotting Cluster Centers')

    clusters_dir = os.path.join(workdir, 'clusters')

    for n_clusters in [4, 7, 10, 13]:
        # count number of files to plot
        centers_files = [os.path.join(clusters_dir, f'cluster_centers_{i}.parquet') for i in range(2, n_clusters + 1)]
        n_files = len(centers_files)
        n_cols = min(n_files, max_cols)
        n_rows = math.ceil(n_files / n_cols)

        # initialize the figure and labels
        fig, axes = plt.subplots(
            n_rows,
            n_cols,
            figsize=(plt_width * n_cols + 1.25, plt_height * n_rows + 1.25),
            dpi=750,
            squeeze=False,
            tight_layout=True,
            sharey='row',
            sharex='col'
        )

        fig.suptitle('Cluster Centers', fontsize=16)
        fig.supylabel('Discharge Z-Score')
        fig.supxlabel('Exceedance Probability (%)')

        for centers_table, ax in zip(centers_files, fig.axes[:n_files]):
            n_clusters = int(centers_table.split('_')[-1].split('.')[0])
            centers_df = pd.read_parquet(centers_table, engine='fastparquet')

            for i in range(int(n_clusters)):
                ax.plot(centers_df[f'{i}'].values, label=f'Cluster {i + 1}')

            # Plot titles and labels
            ax.set_title(f"k={n_clusters} clusters")
            ax.set_xlim(0, 40)
            ax.set_ylim(-2, 4)

        fig.savefig(os.path.join(clusters_dir, f'figure-cluster-centers-{n_clusters}.png'))
        plt.close(fig)
    return


def plot_fit_metrics(workdir: str, plt_width: int = 4, plt_height: int = 4) -> None:
    """
    Plot the cluster metrics, inertia and silhouette score, vs number of clusters

    Args:
        workdir: path to the project directory
        plt_width: width of each subplot in inches
        plt_height: height of each subplot in inches

    Returns:
        None
    """
    logger.info('Plotting Cluster Fit Metrics')

    clusters_dir = os.path.join(workdir, 'clusters')

    df = read_table(workdir, 'cluster_metrics')
    if os.path.exists(get_table_path(workdir, 'cluster_sscores')):
        df = df.merge(read_table(workdir, 'cluster_sscores'), on='number', how='outer')
    df['number'] = df['number'].astype(int)
    df['inertia'] = df['inertia'].astype(float)

    # initialize the figure and labels
    fig, ax = plt.subplots(
        figsize=(plt_width, plt_height),
        dpi=750,
        tight_layout=True,
    )

    # Plot titles and labels
    ax.set_title("Clustering Fit Metrics")
    ax.set_xlabel("Number of Clusters")
    ax.set_ylabel("Inertia")

    ticks = np.arange(1, df['number'].max() + 2)
    ax.set_xlim(ticks[0], ticks[-1])
    ax.set_xticks(ticks)
    ax.set_yticks([])

    # plot the inertia
    knee = int(df['knee'].values[0])
    ax.plot(df['number'], df['inertia'], marker='o', label='Inertia')
    ax.plot(knee, df[df['number'] == knee]['inertia'], marker='o', c='red', label='Knee')

    # add the silhouette scores if they were calculated
    if 'silhouette' in df.columns:
        df.loc[df['silhouette'].isna(), 'silhouette'] = ''
        ax2 = ax.twinx()
        ax2.set_ylabel("Silhouette Score")
        ax2.set_ylim(0, 1)
        ax2.plot(df['number'], df['silhouette'], marker='o', c='green', label='Silhouette Score')

    fig.savefig(os.path.join(clusters_dir, f'figure-fit-metrics.png'))
    plt.close(fig)
    return



import glob
import os
from collections.abc import Iterable
from typing import List

import pandas as pd
from natsort import natsorted

# assign table and gis_input file required column names
mid_col = 'model_id'
gid_col = 'gauge_id'
clbl_col = 'cluster_label'
asgn_mid_col = 'assigned_model_id'
asgn_gid_col = 'assigned_gauge_id'
down_mid_col = 'downstream_model_id'
reason_col = 'reason'
area_col = 'drain_area'
order_col = 'strahler_order'
x_col = 'x_mod'
y_col = 'y_mod'

# dataframe columns names
q_obs = 'Qobs'
q_mod = 'Qmod'
q_sim = 'Qsim'

# name of some files produced by the algorithm
cluster_count_file = 'best-fit-cluster-count.json'
cal_nc_name = 'calibrated_simulated_flow.nc'

# scaffolded folders
dir_tables = 'tables'
dir_gis = 'gis'
dir_clusters = 'clusters'
dir_valid = 'validation'

# name of the required input tables and the outputs
table_hindcast = 'hindcast.parquet'
table_hindcast_fdc = 'hindcast_fdc.parquet'
table_hindcast_fdc_trans = 'hindcast_fdc_transformed.parquet'

table_assign = 'assign_table.parquet'
table_mids = 'mid_table.parquet'
table_gids = 'gid_table.parquet'
table_mid_gid_map = 'mid_gid_map_table.parquet'

table_drain = 'drain_table.parquet'
table_gauge = 'gauge_table.parquet'

table_prop_resolved = 'prop_table_resolved.parquet'
table_prop_downstream = 'prop_table_downstream.parquet'
table_prop_upstream = 'prop_table_upstream.parquet'

# tables generated by the clustering functions
table_cluster_metrics = 'cluster_metrics.csv'
table_cluster_sscores = 'cluster_sscores.csv'
table_cluster_labels = 'cluster_labels.parquet'

# metrics computed on validation sets
metric_list = ['ME', 'MAE', 'RMSE', 'NSE', 'KGE (2012)']
metric_nc_name_list = ['ME', 'MAE', 'RMSE', 'NSE', 'KGE2012']


def scaffold_workdir(path: str, include_validation: bool = True) -> None:
    """
    Creates the correct directories for a Saber project within the specified directory

    Args:
        path: the path to a directory where you want to create workdir subdirectories
        include_validation: boolean, indicates whether to create the validation folder

    Returns:
        None
    """
    dir_list = [dir_tables, dir_gis, dir_clusters]
    if not os.path.exists(path):
        os.mkdir(path)
    if include_validation:
        dir_list.append(dir_valid)
    for d in dir_list:
        p = os.path.join(path, d)
        if not os.path.exists(p):
            os.mkdir(p)
    return


def get_table_path(workdir: str, table_name: str) -> str:
    """
    Get the path to a table in the project directory by name

    Args:
        workdir: path to the project directory
        table_name: name of the table to find a path for

    Returns:
        Path (str) to the table

    Raises:
        ValueError: if the table name is not recognized
    """
    if table_name == 'hindcast':
        return os.path.join(workdir, dir_tables, table_hindcast)
    elif table_name == 'hindcast_fdc':
        return os.path.join(workdir, dir_tables, table_hindcast_fdc)
    elif table_name == 'hindcast_fdc_trans':
        return os.path.join(workdir, dir_tables, table_hindcast_fdc_trans)

    elif table_name == 'assign_table':
        return os.path.join(workdir, dir_tables, table_assign)
    elif table_name == 'drain_table':
        return os.path.join(workdir, dir_tables, table_drain)
    elif table_name == 'gauge_table':
        return os.path.join(workdir, dir_tables, table_gauge)
    elif table_name == 'mid_list':
        return os.path.join(workdir, dir_tables, table_mids)
    elif table_name == 'gid_list':
        return os.path.join(workdir, dir_tables, table_gids)
    elif table_name == 'mid_gid_map':
        return os.path.join(workdir, dir_tables, table_mid_gid_map)

    elif table_name == 'prop_resolved':
        return os.path.join(workdir, dir_tables, table_prop_resolved)
    elif table_name == 'prop_downstream':
        return os.path.join(workdir, dir_tables, table_prop_downstream)
    elif table_name == 'prop_upstream':
        return os.path.join(workdir, dir_tables, table_prop_upstream)

    elif table_name == 'cluster_metrics':
        return os.path.join(workdir, dir_clusters, table_cluster_metrics)
    elif table_name == 'cluster_sscores':
        return os.path.join(workdir, dir_clusters, table_cluster_sscores)
    elif table_name == 'cluster_labels':
        return os.path.join(workdir, dir_clusters, table_cluster_labels)
    elif table_name.startswith('cluster_centers_'):  # cluster_centers_{n_clusters}.parquet - 1 per cluster
        return os.path.join(workdir, dir_clusters, f'{table_name}.parquet')
    elif table_name.startswith('cluster_sscores_'):  # cluster_sscores_{n_clusters}.parquet - 1 per cluster
        return os.path.join(workdir, dir_clusters, f'{table_name}.parquet')

    else:
        raise ValueError(f'Unknown table name: {table_name}')


def read_table(workdir: str, table_name: str) -> pd.DataFrame:
    """
    Read a table from the project directory by name.

    Args:
        workdir: path to the project directory
        table_name: name of the table to read

    Returns:
        pd.DataFrame

    Raises:
        FileNotFoundError: if the table does not exist in the correct directory with the correct name
        ValueError: if the table format is not recognized
    """
    table_path = get_table_path(workdir, table_name)
    if not os.path.exists(table_path):
        raise FileNotFoundError(f'Table does not exist: {table_path}')

    table_format = os.path.splitext(table_path)[-1]
    if table_format == '.parquet':
        return pd.read_parquet(table_path, engine='fastparquet')
    elif table_format == '.feather':
        return pd.read_feather(table_path)
    elif table_format == '.csv':
        return pd.read_csv(table_path, dtype=str)
    else:
        raise ValueError(f'Unknown table format: {table_format}')


def write_table(table: pd.DataFrame, workdir: str, table_name: str) -> None:
    """
    Write a table to the correct location in the project directory

    Args:
        table: the pandas DataFrame to write
        workdir: the path to the project directory
        table_name: the name of the table to write

    Returns:
        None

    Raises:
        ValueError: if the table format is not recognized
    """
    table_path = get_table_path(workdir, table_name)
    table_format = os.path.splitext(table_path)[-1]
    if table_format == '.parquet':
        return table.to_parquet(table_path)
    elif table_format == '.feather':
        return table.to_feather(table_path)
    elif table_format == '.csv':
        return table.to_csv(table_path, index=False)
    else:
        raise ValueError(f'Unknown table format: {table_format}')


def _find_model_files(workdir: str, n_clusters: int or Iterable = 'all') -> List[str]:
    """
    Find all the kmeans model files in the project directory.

    Args:
        workdir: path to the project directory

    Returns:
        List of paths to the kmeans model files

    Raises:
        TypeError: if n_clusters is not an int, iterable of int, or 'all'
    """
    kmeans_dir = os.path.join(workdir, dir_clusters)
    if n_clusters == 'all':
        return natsorted(glob.glob(os.path.join(kmeans_dir, 'kmeans-*.pickle')))
    elif isinstance(n_clusters, int):
        return glob.glob(os.path.join(kmeans_dir, f'kmeans-{n_clusters}.pickle'))
    elif isinstance(n_clusters, Iterable):
        return natsorted([os.path.join(kmeans_dir, f'kmeans-{i}.pickle') for i in n_clusters])
    else:
        raise TypeError('n_clusters should be of type int or an iterable')



import logging
import os
import warnings
from multiprocessing import Pool

import hydrostats as hs
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from sklearn.model_selection import RepeatedKFold as RKFolds

from .assign import map_assign_ungauged
from .assign import mp_assign_all
from .calibrate import map_saber, mp_saber
from .io import asgn_gid_col
from .io import asgn_mid_col
from .io import dir_valid
from .io import gid_col
from .io import mid_col
from .io import q_mod
from .io import q_obs
from .io import q_sim

__all__ = ['kfolds', 'bootstrap', 'mp_bootstrap', 'bootstrap_figures']

logger = logging.getLogger(__name__)

warnings.filterwarnings('ignore')


def kfolds(workdir: str, assign_df: pd.DataFrame, gauge_data: str, hds: str,
           n_splits: int = 5, n_repeats: int = 10, n_processes: int or None = None) -> None:
    """
    Performs a k-fold validation of the calibration results. The validation set is randomly selected from the
    gauges table. The validation set is then removed from the calibration set and the calibration is repeated
    n_repeats times. The validation set is then added back to the calibration set and the calibration is repeated
    n_repeats times. This process is repeated n_folds times.

    Args:
        workdir: the project working directory
        assign_df: a completed assignment dataframe already filled by a successful SABER assignment run
        gauge_data: path to the directory of observed data
        hds: string path to the hindcast streamflow dataset
        n_splits: the number of folds to create in each iteration
        n_repeats: the number of repeats to perform
        n_processes: number of processes to use for multiprocessing, passed to Pool

    Returns:
        pandas.DataFrame
    """
    # subset the assign dataframe to only rows which contain gauges & reset the index
    assign_df = assign_df[assign_df[gid_col].notna()].reset_index(drop=True)

    for i, (train_rows, test_rows) in enumerate(RKFolds(n_splits=n_splits, n_repeats=n_repeats).split(assign_df.index)):
        # make a copy of the dataframe to work with for this iteration
        val_df = assign_df.copy()

        # on the test rows, clear the gid column so that it is not assigned to the gauge it contains
        val_df.loc[test_rows, gid_col] = np.nan

        # make assignments on the training set
        val_df = mp_assign_all(workdir, val_df, n_processes=n_processes)

        # reduce the dataframe to only the test rows (index/row-order gets shuffled by mp assigning)
        val_df = val_df[val_df[gid_col].isna()].reset_index(drop=True)

        # bias correct only at the gauges in the test set
        mp_saber(val_df, hds, gauge_data, os.path.join(workdir, dir_valid), n_processes=n_processes)

        # todo calculate metrics for the test set
    return


def bootstrap(row_idx: int, assign_df: pd.DataFrame, gauge_data: str, hds: str,
              save_dir: str) -> pd.DataFrame | None:
    """
    Performs bootstrap validation.

    Args:
        row_idx: the row of the assignment table to remove and perform bootstrap validation with
        assign_df: pandas.DataFrame of the assignment table
        gauge_data: string path to the directory of observed data
        hds: string path to the hindcast streamflow dataset
        save_dir: string path to the directory to save the corrected data

    Returns:
        None
    """
    try:
        row = assign_df.loc[row_idx]
        assigned_row = map_assign_ungauged(assign_df, assign_df.drop(row_idx), row[mid_col])
        corrected_df = map_saber(
            assigned_row[mid_col].values[0],
            assigned_row[asgn_mid_col].values[0],
            assigned_row[asgn_gid_col].values[0],
            hds,
            gauge_data,
            save_dir
        )

        if corrected_df is None:
            logger.warning(f'No corrected data for {row[mid_col]}')
            return None
        if not all([q_mod in corrected_df.columns, q_sim in corrected_df]):
            logger.warning(f'missing columns')
            return None

        metrics_df = pd.read_csv(os.path.join(gauge_data, f'{row[gid_col]}.csv'), index_col=0)
        metrics_df.columns = [q_obs, ]
        metrics_df.index = pd.to_datetime(metrics_df.index)
        metrics_df = pd.merge(corrected_df, metrics_df, how='inner', left_index=True, right_index=True)

        obs_values = metrics_df[q_obs].values.flatten()
        sim_values = metrics_df[q_sim].values.flatten()
        mod_values = np.squeeze(metrics_df[q_mod].values.flatten())

        if mod_values.dtype == np.dtype('O'):
            mod_values = np.array(mod_values.tolist()).astype(np.float64).flatten()

        diff_sim = sim_values - obs_values
        diff_corr = mod_values - obs_values

        return pd.DataFrame({
            'me_sim': np.mean(diff_sim),
            'mae_sim': np.mean(np.abs(diff_sim)),
            'rmse_sim': np.sqrt(np.mean(diff_sim ** 2)),
            'nse_sim': hs.nse(sim_values, obs_values),
            'kge_sim': hs.kge_2012(sim_values, obs_values),

            'me_corr': np.mean(diff_corr),
            'mae_corr': np.mean(np.abs(diff_corr)),
            'rmse_corr': np.sqrt(np.mean(diff_corr ** 2)),
            'nse_corr': hs.nse(mod_values, obs_values),
            'kge_corr': hs.kge_2012(mod_values, sim_values),

            'reach_id': row[mid_col],
            'gauge_id': row[gid_col],
            'asgn_reach_id': assigned_row[asgn_mid_col],
        })
    except Exception as e:
        logger.error(e)
        logger.error(f'Failed bootstrap validation for {row[mid_col]}')
        return None


def mp_bootstrap(workdir: str, assign_df: pd.DataFrame, gauge_data: str, hds: str, n_processes: int = 1) -> None:
    """
    Performs bootstrap validation using multiprocessing.

    Args:
        workdir: the project working directory
        assign_df: pandas.DataFrame of the assignment table
        gauge_data: string path to the directory of observed data
        hds: string path to the hindcast streamflow dataset
        n_processes: number of processes to use for multiprocessing, passed to Pool

    Returns:
        None
    """
    # subset the assign dataframe to only rows which contain gauges & reset the index
    assign_df = assign_df[assign_df[gid_col].notna()].reset_index(drop=True)
    save_dir = os.path.join(workdir, dir_valid)

    with Pool(n_processes) as p:
        metrics_df = pd.concat(
            p.starmap(
                bootstrap,
                [[idx, assign_df, gauge_data, hds, save_dir] for idx in assign_df.index]
            )
        )

    metrics_df.to_csv(os.path.join(save_dir, 'bootstrap_metrics.csv'), index=False)

    return


def bootstrap_figures(workdir: str, df: pd.DataFrame = None, stat: str = None) -> None:
    """

    Args:
        workdir: the project working directory
        df: pandas.DataFrame of the bootstrap metrics
        stat: the statistic to plot

    Returns:
        None
    """
    if df is None:
        df = pd.read_csv(os.path.join(workdir, dir_valid, 'bootstrap_metrics.csv'), index_col=0)

    if stat is None:
        for stat in ['me', 'mae', 'rmse', 'nse', 'kge']:
            bootstrap_figures(workdir, df, stat)
        return

    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(8, 4), dpi=1500, tight_layout=True, sharey=True)

    if stat == 'kge':
        binwidth = 0.5
        binrange = (-6, 1)
        ax1.axvline(-0.44, c='red', linestyle='--')
        ax2.axvline(-0.44, c='red', linestyle='--')

    elif stat == 'nse':
        binwidth = 0.5
        binrange = (-6, 1)

    elif stat == 'me':
        binwidth = 20
        binrange = (-200, 200)

    elif stat == 'mae':
        binwidth = 30
        binrange = (0, 300)

    elif stat == 'rmse':
        binwidth = 20
        binrange = (0, 200)

    else:
        raise ValueError(f'Invalid statistic: {stat}')

    fig.suptitle(f'Bootstrap Validation: {stat.upper()}')
    ax1.grid(True, 'both', zorder=0, linestyle='--')
    ax2.grid(True, 'both', zorder=0, linestyle='--')
    ax1.set_xlim(binrange)
    ax2.set_xlim(binrange)

    stat_df = df[[f'{stat}_corr', f'{stat}_sim']].reset_index(drop=True).copy()
    stat_df[stat_df <= binrange[0]] = binrange[0]
    stat_df[stat_df >= binrange[1]] = binrange[1]

    sns.histplot(stat_df, x=f'{stat}_sim', binwidth=binwidth, binrange=binrange, ax=ax1)
    sns.histplot(stat_df, x=f'{stat}_corr', binwidth=binwidth, binrange=binrange, ax=ax2)

    ax1.axvline(stat_df[f'{stat}_sim'].median(), c='green')
    ax2.axvline(stat_df[f'{stat}_corr'].median(), c='green')

    fig.savefig(os.path.join(workdir, dir_valid, f'bootstrap_{stat}.png'))
    return



import logging
import os
import statistics
from multiprocessing import Pool

import numpy as np
import pandas as pd
import xarray
from natsort import natsorted
from scipy import interpolate, stats

from .io import asgn_mid_col
from .io import gid_col
from .io import mid_col
from .io import q_mod
from .io import q_obs
from .io import q_sim

logger = logging.getLogger(__name__)

__all__ = ['mp_saber', 'fdc_mapping', 'sfdc_mapping', 'map_saber', 'calc_fdc', 'calc_sfdc']


def mp_saber(assign_df: pd.DataFrame, hds: str, gauge_data: str, save_dir: str = None,
             n_processes: int or None = None) -> None:
    """
    Corrects all streams in the assignment table using the SABER method with a multiprocessing Pool

    Args:
        assign_df: the assignment table
        hds: string path to the hindcast streamflow dataset
        gauge_data: path to the directory of observed data
        save_dir: path to the directory to save the corrected data
        n_processes: number of processes to use for multiprocessing, passed to Pool

    Returns:
        None
    """
    logger.info('Starting SABER Bias Correction')

    if save_dir is None:
        save_dir = os.path.join(gauge_data, 'corrected')
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)

    with Pool(n_processes) as p:
        p.starmap(
            map_saber,
            [[mid, asgn_mid, asgn_gid, hds, gauge_data, save_dir] for mid, asgn_mid, asgn_gid in
             np.moveaxis(assign_df[[mid_col, asgn_mid_col, gid_col]].values, 0, 0)]
        )

    logger.info('Finished SABER Bias Correction')
    return


def map_saber(mid: str, asgn_mid: str, asgn_gid: str, hds: str, gauge_data: str,
              save_dir: str) -> pd.DataFrame | tuple | None:
    """
    Corrects all streams in the assignment table using the SABER method

    Args:
        mid: the model id of the stream to be corrected
        asgn_mid: the model id of the stream assigned to mid for bias correction
        asgn_gid: the gauge id of the stream assigned to mid for bias correction
        hds: xarray dataset of hindcast streamflow data
        gauge_data: path to the directory of observed data
        save_dir: path to the directory to save the corrected data

    Returns:
        None
    """
    try:
        if asgn_gid is None or pd.isna(asgn_gid):
            logger.debug(f'No gauge assigned to {mid}')
            return

        # find the observed data to be used for correction
        if not os.path.exists(os.path.join(gauge_data, f'{asgn_gid}.csv')):
            logger.debug(f'Observed data "{asgn_gid}" not found. Cannot correct "{mid}".')
        obs_df = pd.read_csv(os.path.join(gauge_data, f'{asgn_gid}.csv'), index_col=0)
        obs_df.index = pd.to_datetime(obs_df.index)

        # perform corrections
        hds = xarray.open_mfdataset(hds, concat_dim='rivid', combine='nested', parallel=True, engine='zarr')
        rivids = hds.rivid.values
        sim_a = hds['Qout'][:, rivids == int(mid)].values
        sim_a = pd.DataFrame(sim_a, index=hds['time'].values, columns=[q_sim])
        if asgn_mid != mid:
            sim_b = hds['Qout'][:, rivids == int(asgn_mid)].values
            sim_b = pd.DataFrame(sim_b, index=sim_a.index, columns=[q_sim])
        hds.close()

        if asgn_mid == mid:
            corrected_df = fdc_mapping(sim_a, obs_df)
        else:
            corrected_df = sfdc_mapping(
                sim_b, obs_df, sim_a,
                drop_outliers=True, outlier_threshold=3,
                fit_gumbel=True, fit_range=(5, 95),
            )

        # save the corrected data
        corrected_df.to_csv(os.path.join(save_dir, f'saber_{mid}.csv'))
        return corrected_df

    except Exception as e:
        logger.error(e)
        logger.debug(f'Failed to correct {mid}')
        return


def fdc_mapping(sim_df: pd.DataFrame, obs_df: pd.DataFrame) -> pd.DataFrame:
    """
    Bias corrects a dataframe of simulated values using a dataframe of observed values

    Args:
        sim_df: A dataframe with a datetime index and a single column of streamflow values
        obs_df: A dataframe with a datetime index and a single column of streamflow values
    Returns:
        pandas DataFrame with a datetime index and a single column of streamflow values
    """
    dates = []
    values = []

    for month in natsorted(sim_df.index.month.unique()):
        # filter historical data to only be current month
        month_sim = sim_df[sim_df.index.month == int(month)].dropna()
        month_obs = obs_df[obs_df.index.month == int(month)].dropna()

        # calculate the flow duration curves
        month_sim_fdc = calc_fdc(month_sim.values)
        month_obs_fdc = calc_fdc(month_obs.values)

        # make interpolator for 1) sim flow to sim prob, and 2) obs prob to obs flow
        to_prob = _make_interpolator(month_sim_fdc.values.flatten(), month_sim_fdc.index)
        to_flow = _make_interpolator(month_obs_fdc.index, month_obs_fdc.values.flatten())

        dates += month_sim.index.to_list()
        values += to_flow(to_prob(month_sim.values)).tolist()

    return pd.DataFrame({
        q_mod: values,
        q_sim: sim_df.values.flatten(),
    }, index=dates).sort_index()


def sfdc_mapping(sim_flow_a: pd.DataFrame, obs_flow_a: pd.DataFrame, sim_flow_b: pd.DataFrame = None,
                 fix_seasonally: bool = True, empty_months: str = 'skip',
                 drop_outliers: bool = False, outlier_threshold: int or float = 2.5,
                 filter_scalar_fdc: bool = False, filter_range: tuple = (0, 80),
                 extrapolate: str = 'nearest', fill_value: int or float = None,
                 fit_gumbel: bool = False, fit_range: tuple = (10, 90),
                 metadata: bool = False, ) -> pd.DataFrame:
    """
    Removes the bias from simulated discharge using the SABER method.

    Given simulated and observed discharge at location A, removes bias from simulated data at point A.
    Given simulated and observed discharge at location A, removes bias from simulated data at point B, if given B.

    Args:
        sim_flow_a (pd.DataFrame): simulated hydrograph at point A. should contain a datetime index with daily values
            and a single column of discharge values.
        obs_flow_a (pd.DataFrame): observed hydrograph at point A. should contain a datetime index with daily values
            and a single column of discharge values.
        sim_flow_b (pd.DataFrame): (optional) simulated hydrograph at point B to correct using scalar flow duration
            curve mapping and the bias relationship at point A. should contain a datetime index with daily values
            and a single column of discharge values.

        fix_seasonally (bool): fix on a monthly (True) or annual (False) basis
        empty_months (str): how to handle months in the simulated data where no observed data are available. Options:
            "skip": ignore simulated data for months without

        drop_outliers (bool): flag to exclude outliers
        outlier_threshold (int or float): number of std deviations from mean to exclude from flow duration curve

        filter_scalar_fdc (bool): flag to filter the scalar flow duration curve
        filter_range (tuple): lower and upper bounds of the filter range

        extrapolate (str): method to use for extrapolation. Options: nearest, const, linear, average, max, min
        fill_value (int or float): value to use for extrapolation when extrapolate_method='const'

        fit_gumbel (bool): flag to replace extremely low/high corrected flows with values from Gumbel type 1
        fit_range (tuple): lower and upper bounds of exceedance probabilities to replace with Gumbel values

        metadata (bool): flag to return the scalars and metadata about the correction process

    Returns:
        pd.DataFrame with a DateTime index and columns with corrected flow, uncorrected flow, the scalar adjustment
        factor applied to correct the discharge, and the percentile of the uncorrected flow (in the seasonal grouping,
        if applicable).
    """
    if fix_seasonally:
        # list of the unique months in the historical simulation. should always be 1->12 but just in case...
        monthly_results = []
        for month in sorted(set(sim_flow_a.index.strftime('%m'))):
            # filter data to current iteration's month
            mon_obs_a = obs_flow_a[obs_flow_a.index.month == int(month)].dropna()

            if mon_obs_a.empty:
                if empty_months == 'skip':
                    continue
                else:
                    raise ValueError(f'Invalid value for argument "empty_months". Given: {empty_months}.')

            mon_sim_a = sim_flow_a[sim_flow_a.index.month == int(month)].dropna()
            mon_sim_b = sim_flow_b[sim_flow_b.index.month == int(month)].dropna()
            monthly_results.append(sfdc_mapping(
                mon_sim_a, mon_obs_a, mon_sim_b,
                fix_seasonally=False, empty_months=empty_months,
                drop_outliers=drop_outliers, outlier_threshold=outlier_threshold,
                filter_scalar_fdc=filter_scalar_fdc, filter_range=filter_range,
                extrapolate=extrapolate, fill_value=fill_value,
                fit_gumbel=fit_gumbel, fit_range=fit_range, )
            )
        # combine the results from each monthly into a single dataframe (sorted chronologically) and return it
        return pd.concat(monthly_results).sort_index()

    # compute the flow duration curves
    if drop_outliers:
        sim_fdc_a = calc_fdc(_drop_outliers_by_zscore(sim_flow_a, threshold=outlier_threshold), col_name=q_sim)
        sim_fdc_b = calc_fdc(_drop_outliers_by_zscore(sim_flow_b, threshold=outlier_threshold), col_name=q_sim)
        obs_fdc = calc_fdc(_drop_outliers_by_zscore(obs_flow_a, threshold=outlier_threshold), col_name=q_obs)
    else:
        sim_fdc_a = calc_fdc(sim_flow_a, col_name=q_sim)
        sim_fdc_b = calc_fdc(sim_flow_b, col_name=q_sim)
        obs_fdc = calc_fdc(obs_flow_a, col_name=q_obs)

    # calculate the scalar flow duration curve (at point A with simulated and observed data)
    scalar_fdc = calc_sfdc(sim_fdc_a[q_sim], obs_fdc[q_obs])
    if filter_scalar_fdc:
        scalar_fdc = scalar_fdc[scalar_fdc['p_exceed'].between(filter_range[0], filter_range[1])]

    # make interpolators: Q_b -> p_exceed_b, p_exceed_a -> scalars_a
    # flow at B converted to exceedance probabilities, then matched with the scalar computed at point A
    flow_to_percent = _make_interpolator(sim_fdc_b.values.flatten(),
                                         sim_fdc_b.index,
                                         extrap=extrapolate,
                                         fill_value=fill_value)

    percent_to_scalar = _make_interpolator(scalar_fdc.index,
                                           scalar_fdc.values.flatten(),
                                           extrap=extrapolate,
                                           fill_value=fill_value)

    # apply interpolators to correct flows at B with data from A
    qb_original = sim_flow_b.values.flatten()
    p_exceed = flow_to_percent(qb_original)
    scalars = percent_to_scalar(p_exceed)
    qb_adjusted = qb_original / scalars

    if fit_gumbel:
        qb_adjusted = _fit_extreme_values_to_gumbel(qb_adjusted, p_exceed, fit_range)

    response = pd.DataFrame(data=np.transpose([qb_adjusted, qb_original]),
                            index=sim_flow_b.index.to_list(),
                            columns=(q_mod, q_sim))
    if metadata:
        response['scalars'] = scalars
        response['p_exceed'] = p_exceed

    return response


def calc_fdc(flows: np.array, steps: int = 201, col_name: str = 'Q') -> pd.DataFrame:
    """
    Compute flow duration curve (exceedance probabilities) from a list of flows

    Args:
        flows: array of flows
        steps: number of steps (exceedance probabilities) to use in the FDC
        col_name: name of the column in the returned dataframe

    Returns:
        pd.DataFrame with index 'p_exceed' and columns 'Q' (or col_name)
    """
    # calculate the FDC and save to parquet
    exceed_prob = np.linspace(100, 0, steps)
    fdc_flows = np.nanpercentile(flows, exceed_prob)
    df = pd.DataFrame(fdc_flows, columns=[col_name, ], index=exceed_prob)
    df.index.name = 'p_exceed'
    return df


def calc_sfdc(sim_fdc: pd.DataFrame, obs_fdc: pd.DataFrame) -> pd.DataFrame:
    """
    Compute the scalar flow duration curve (exceedance probabilities) from two flow duration curves

    Args:
        sim_fdc: simulated flow duration curve
        obs_fdc: observed flow duration curve

    Returns:
        pd.DataFrame with index (exceedance probabilities) and a column of scalars
    """
    scalars_df = pd.DataFrame(
        np.divide(sim_fdc.values.flatten(), obs_fdc.values.flatten()),
        columns=['scalars'],
        index=sim_fdc.index
    )
    scalars_df.replace(np.inf, np.nan, inplace=True)
    scalars_df.dropna(inplace=True)
    return scalars_df


def _drop_outliers_by_zscore(df: pd.DataFrame, threshold: float = 3) -> pd.DataFrame:
    """
    Drop outliers from a dataframe by their z-score and a threshold
    Based on https://stackoverflow.com/questions/23199796/detect-and-exclude-outliers-in-pandas-data-frame
    Args:
        df: dataframe to drop outliers from
        threshold: z-score threshold

    Returns:
        pd.DataFrame with outliers removed
    """
    return df[(np.abs(stats.zscore(df)) < threshold).all(axis=1)]


def _filter_sfdc(sfdc: pd.DataFrame, filter_range: list) -> pd.DataFrame:
    """
    Filter the scalar flow duration curve by the specified range

    Args:
        sfdc: scalar flow duration curve DataFrame
        filter_range: list of [lower_bound: int or float, upper_bound: int or float]

    Returns:
        pd.DataFrame: filtered scalar flow duration curve
    """
    return sfdc[np.logical_and(sfdc.index > filter_range[0], sfdc.index < filter_range[1])]


def _make_interpolator(x: np.array, y: np.array, extrap: str = 'nearest',
                       fill_value: int or float = None) -> interpolate.interp1d:
    """
    Make an interpolator from two arrays

    Args:
        x: x values
        y: y values
        extrap: method for extrapolation: nearest, const, linear, average, max, min
        fill_value: value to use when extrap='const'

    Returns:
        interpolate.interp1d
    """
    # todo check that flows are not negative and have sufficient variance - even for small variance in SAF
    # if np.max(y) - np.min(y) < 5:
    #     logger.warning('The y data has similar max/min values. You may get unanticipated results.')

    # make interpolator which converts the percentiles to scalars
    if extrap == 'nearest':
        return interpolate.interp1d(x, y, fill_value='extrapolate', kind='nearest')
    elif extrap == 'const':
        if fill_value is None:
            raise ValueError('Must provide the const kwarg when extrap_method="const"')
        return interpolate.interp1d(x, y, fill_value=fill_value, bounds_error=False)
    elif extrap == 'linear':
        return interpolate.interp1d(x, y, fill_value='extrapolate')
    elif extrap == 'average':
        return interpolate.interp1d(x, y, fill_value=np.mean(y), bounds_error=False)
    elif extrap == 'max' or extrap == 'maximum':
        return interpolate.interp1d(x, y, fill_value=np.max(y), bounds_error=False)
    elif extrap == 'min' or extrap == 'minimum':
        return interpolate.interp1d(x, y, fill_value=np.min(y), bounds_error=False)
    else:
        raise ValueError('Invalid extrapolation method provided')


def _solve_gumbel1(std, xbar, rp):
    """
    Solves the Gumbel Type I pdf = exp(-exp(-b))

    Args:
        std: standard deviation of the dataset
        xbar: mean of the dataset
        rp: return period to calculate in years

    Returns:
        float: discharge value
    """
    return -np.log(-np.log(1 - (1 / rp))) * std * .7797 + xbar - (.45 * std)


def _fit_extreme_values_to_gumbel(q_adjust: np.array, p_exceed: np.array, fit_range: tuple = None) -> np.array:
    """
    Replace the extreme values from the corrected data with values based on the gumbel distribution

    Args:
        q_adjust: adjusted flows to be refined
        p_exceed: exceedance probabilities of the adjusted flows
        fit_range: range of exceedance probabilities to fit to the Gumbel distribution

    Returns:
        array of the flows with the extreme values replaced
    """
    all_values = pd.DataFrame(np.transpose([q_adjust, p_exceed]), columns=('q', 'p'))

    # compute the average and standard deviation for the values within the user specified fit_range
    mid_vals = all_values[np.logical_and(all_values['p'] >= fit_range[0], all_values['p'] <= fit_range[1])]
    xbar = statistics.mean(mid_vals['q'].values)
    std = statistics.stdev(mid_vals['q'].values, xbar)

    outlier_vals = all_values.drop(mid_vals.index)
    outlier_vals['q'] = outlier_vals['q'] = -np.log(
        -np.log(1 - (1 / (1 / (1 - (outlier_vals['p'] / 100)))))) * std * .7797 + xbar - (.45 * std)
    outlier_vals[outlier_vals < 0] = 0
    all_values.update(outlier_vals)

    return all_values['q'].values.flatten()



import os
import logging

import contextily as cx
import geopandas as gpd
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from .io import gid_col
from .io import metric_nc_name_list
from .io import mid_col
from .io import reason_col
from .io import dir_gis
from .io import clbl_col

__all__ = ['create_maps', 'map_by_reason', 'map_by_cluster', 'map_unassigned', 'map_ids',
           'validation_maps']

logger = logging.getLogger(__name__)


def create_maps(workdir: str, assign_df: pd.DataFrame, drain_gis: str or gpd.GeoDataFrame, prefix: str = '') -> None:
    """
    Runs all the clip functions which create subsets of the drainage lines GIS dataset based on how they were assigned
    for bias correction.

    Args:
        workdir: the path to the working directory for the project
        assign_df: the assignment table dataframe
        drain_gis: path to a drainage line shapefile which can be clipped
        prefix: a prefix for names of the outputs to distinguish between data generated in separate instances

    Returns:
        None
    """
    if type(drain_gis) == str:
        gdf = gpd.read_file(drain_gis)
    elif type(drain_gis) == gpd.GeoDataFrame:
        gdf = drain_gis
    else:
        raise TypeError(f'Invalid type for drain_gis: {type(drain_gis)}')

    map_by_reason(workdir, assign_df, gdf, prefix)
    map_by_cluster(workdir, assign_df, gdf, prefix)
    map_unassigned(workdir, assign_df, gdf, prefix)
    return


def map_by_reason(workdir: str, assign_df: pd.DataFrame, drain_gis: str or gpd.GeoDataFrame, prefix: str = '') -> None:
    """
    Creates Geopackage files in workdir/gis_outputs for each unique value in the assignment column

    Args:
        workdir: the path to the working directory for the project
        assign_df: the assignment table dataframe
        drain_gis: path to a drainage line shapefile which can be clipped
        prefix: a prefix for names of the outputs to distinguish between data generated at separate instances

    Returns:
        None
    """
    # read the drainage line shapefile
    if isinstance(drain_gis, str):
        drain_gis = gpd.read_file(drain_gis)

    # get the unique list of assignment reasons
    for reason in assign_df[reason_col].unique():
        logger.info(f'Creating GIS output for group: {reason}')
        selector = drain_gis[mid_col].astype(str).isin(assign_df[assign_df[reason_col] == reason][mid_col])
        subset = drain_gis[selector]
        name = f'{f"{prefix}_" if prefix else ""}assignments_{reason}.gpkg'
        if subset.empty:
            logger.debug(f'Empty filter: No streams are assigned for {reason}')
            continue
        else:
            subset.to_file(os.path.join(workdir, dir_gis, name))
    return


def map_by_cluster(workdir: str, assign_table: pd.DataFrame, drain_gis: str, prefix: str = '') -> None:
    """
    Creates Geopackage files in workdir/gis_outputs of the drainage lines based on the fdc cluster they were assigned to

    Args:
        workdir: the path to the working directory for the project
        assign_table: the assignment table dataframe
        drain_gis: path to a drainage line shapefile which can be clipped
        prefix: optional, a prefix to prepend to each created file's name

    Returns:
        None
    """
    if isinstance(drain_gis, str):
        drain_gis = gpd.read_file(drain_gis)
    for num in assign_table[clbl_col].unique():
        logger.info(f'Creating GIS output for cluster: {num}')
        gdf = drain_gis[drain_gis[mid_col].astype(str).isin(assign_table[assign_table[clbl_col] == num][mid_col])]
        if gdf.empty:
            logger.debug(f'Empty filter: No streams are assigned to cluster {num}')
            continue
        gdf.to_file(os.path.join(workdir, dir_gis, f'{prefix}{"_" if prefix else ""}cluster-{int(num)}.gpkg'))
    return


def map_unassigned(workdir: str, assign_table: pd.DataFrame, drain_gis: str, prefix: str = '') -> None:
    """
    Creates Geopackage files in workdir/gis_outputs of the drainage lines which haven't been assigned a gauge yet

    Args:
        workdir: the path to the working directory for the project
        assign_table: the assignment table dataframe
        drain_gis: path to a drainage line shapefile which can be clipped
        prefix: optional, a prefix to prepend to each created file's name

    Returns:
        None
    """
    logger.info('Creating GIS output for unassigned basins')
    if isinstance(drain_gis, str):
        drain_gis = gpd.read_file(drain_gis)
    ids = assign_table[assign_table[reason_col] == 'unassigned'][mid_col].values
    subset = drain_gis[drain_gis[mid_col].astype(str).isin(ids)]
    if subset.empty:
        logger.debug('Empty filter: No streams are unassigned')
        return
    savepath = os.path.join(workdir, dir_gis, f'{prefix}{"_" if prefix else ""}assignments_unassigned.gpkg')
    subset.to_file(savepath)
    return


def map_ids(workdir: str, ids: list, drain_gis: str, prefix: str = '', id_column: str = mid_col) -> None:
    """
    Creates Geopackage files in workdir/gis_outputs of the subset of 'drain_shape' with an ID in the specified list

    Args:
        workdir: path to the project directory
        ids: any iterable containing a series of model_ids
        drain_gis: path to the drainage shapefile to be clipped
        prefix: optional, a prefix to prepend to each created file's name
        id_column: name of the id column in the attributes of the shape table

    Returns:
        None
    """
    if isinstance(drain_gis, str):
        drain_gis = gpd.read_file(drain_gis)
    name = f'{prefix}{"_" if prefix else ""}id_subset.gpkg'
    drain_gis[drain_gis[id_column].isin(ids)].to_file(os.path.join(workdir, dir_gis, name))
    return


def validation_maps(workdir: str, gauge_gis: str, val_table: pd.DataFrame = None, prefix: str = '') -> None:
    """
    Creates Geopackage files in workdir/gis_outputs of subsets of the gauge_shape.
    1 is the fill gauge shape with added attribute columns for all the computed stats. There are 2 for each of the 5
    validation groups; 1 which shows the gauges included in the validation set and 1 which shows gauges that were
    excluded from the validation set.

    Args:
        workdir: path to the project directory
        val_table: the validation table produced by saber.validate
        gauge_gis: path to the gauge locations shapefile
        prefix: optional, a prefix to prepend to each created file's name

    Returns:
        None
    """
    if val_table is None:
        val_table = pd.read_csv(os.path.join(workdir, 'validation_runs', 'val_table.csv'))
    save_dir = os.path.join(workdir, 'gis')

    # merge gauge table with the validation table
    gdf = gpd.read_file(gauge_gis)
    gdf = gdf.merge(val_table, on=gid_col, how='inner')
    gdf.to_file(os.path.join(save_dir, 'gauges_with_validation_stats.gpkg'))

    core_columns = [mid_col, gid_col, 'geometry']

    # generate gis files by validation run, by stat, and by included/excluded
    for val_set in ('50', '60', '70', '80', '90'):
        for metric in metric_nc_name_list:
            # select only columns for the validation run we're iterating on - too complex for filter/regex
            cols_to_select = core_columns + [val_set, f'{metric}_{val_set}']
            gdf_sub = gdf[cols_to_select]
            gdf_sub = gdf_sub.rename(columns={f'{metric}_{val_set}': metric})

            name = f'{prefix}{"_" if prefix else ""}valset_{val_set}_{metric}_included.gpkg'
            gdf_sub[gdf_sub[val_set] == 1].to_file(os.path.join(save_dir, name))

            name = f'{prefix}{"_" if prefix else ""}valset_{val_set}_{metric}_excluded.gpkg'
            exc = gdf_sub[gdf_sub[val_set] == 0]
            exc.to_file(os.path.join(save_dir, name))
            if metric == 'KGE2012':
                histomaps(exc, metric, val_set, workdir)

    return


def histomaps(gdf: gpd.GeoDataFrame, metric: str, prct: str, workdir: str) -> None:
    """
    Creates a histogram of the KGE2012 values for the validation set

    Args:
        gdf: a GeoDataFrame containing validation metrics
        metric:name of th emetric to plot
        prct: Percentile of the validation set used to generate the histogram
        workdir: the project working directory

    Returns:
        None
    """
    core_columns = [mid_col, gid_col, 'geometry']
    # world = gpd.read_file(gpd.datasets.get_path('naturalearth_lowres'))
    # world.plot(ax=axm, color='white', edgecolor='black')

    colors = ['#dc112e', '#d6db12', '#da9707', '#13c208', '#0824c2']
    bins = [-10, 0, 0.25, 0.5, 0.75, 1]
    cmap = mpl.colors.ListedColormap(colors)
    norm = mpl.colors.BoundaryNorm(boundaries=bins, ncolors=len(cmap.colors))
    title = metric.replace('KGE2012', 'Kling Gupta Efficiency 2012 - ') + f' {prct}% Gauges Excluded'

    hist_groups = []
    hist_colors = []
    categorize_by = [-np.inf, 0, 0.25, 0.5, 0.75, 1]
    for idx in range(len(categorize_by) - 1):
        gdfsub = gdf[gdf[metric] >= categorize_by[idx]]
        gdfsub = gdfsub[gdfsub[metric] < categorize_by[idx + 1]]
        if not gdfsub.empty:
            hist_groups.append(gdfsub[metric].values)
            hist_colors.append(colors[idx])

    fig, (axh, axm) = plt.subplots(
        1, 2, tight_layout=True, figsize=(9, 5), dpi=400, gridspec_kw={'width_ratios': [1, 1]})
    fig.suptitle(title, fontsize=20)

    median = round(gdf[metric].median(), 2)
    axh.set_title(f'Histogram (Median = {median})')
    axh.set_ylabel('Count')
    axh.set_xlabel('KGE 2012')
    axh.hist(hist_groups, color=hist_colors, bins=25, histtype='barstacked', edgecolor='black')
    axh.axvline(median, color='k', linestyle='dashed', linewidth=3)

    axm.set_title('Gauge Map')
    axm.set_ylabel('Latitude')
    axm.set_xlabel('Longitude')
    axm.set_xticks([])
    axm.set_yticks([])
    gdf[core_columns + [metric, ]].to_crs(epsg=3857).plot(metric)
    cx.add_basemap(ax=axm, zoom=9, source=cx.providers.Esri.WorldTopoMap, attribution='')

    fig.savefig(os.path.join(workdir, 'gis', f'{metric}_{prct}.png'))
    return



from setuptools import setup

with open("README.md", "r") as readme:
    long_description = readme.read()

AUTHOR = 'Riley C. Hales'
DESCRIPTION = 'The SABER tool for bias correcting large hydrologic models'

VERSION = '0.7.0'

PYTHON_REQUIRES = '>=3.10'
INSTALL_REQUIRES = [
    'contextily',
    'fastparquet',
    'geopandas',
    'hydrostats',
    'joblib',
    'kneed',
    'matplotlib',
    'natsort',
    'numba',
    'numpy',
    'pandas',
    'pyarrow',
    'requests',
    'scikit-learn',
    'scipy',
    'seaborn',
    'xarray',
    'zarr'
]

TROVE_CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Programming Language :: Python :: 3',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Hydrology',
    'Topic :: Scientific/Engineering :: Visualization',
    'Topic :: Scientific/Engineering :: GIS',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: BSD License',
    'Natural Language :: English',
]

setup(
    name='saber-hbc',
    packages=['saber', ],
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    license='BSD 3-Clause',
    classifiers=TROVE_CLASSIFIERS,
    python_requires=PYTHON_REQUIRES,
    install_requires=INSTALL_REQUIRES
)
